#!/bin/bash
#
# Copyright 2012 Doerzbach Engineering GmbH 
#
# http://www.doerzbach.com
# 
# Written by Armin Doerzbach 
# 

# Source common functions
. `dirname $0`/functions.bash

if [ "$LOGLEVEL" = "DEBUG" ] ; then
  set -xv
fi
MAXRETRIES=5
PATH=/usr/bin:/usr/sbin
CMD=`basename $0`

TMPDIR=/tmp/$CMD.$$
mkdir $TMPDIR

trap "cleanup 100" 1 2 15
cleanup() {
  rm -r $TMPDIR
  exit $1
}
usage() {
  echo "Usage: `basename $0` -s source_zfs -d dest_zfs -c user@host -i interval -p ssh_privatekeyfile [-a afterscript] [-b beforescript] [-r retries] [-k keepsnaps] [[-s source_zfs -d dest_zfs]...] [-C]" >&2
  echo >&2
  echo "The script replicates the source_zfs to the dest_zfs using snapshot send/recieve via ssh with the user and host specified every interval seconds" >&2
  echo "The beforescript is executed before the snapshot is taken and the afterscript after the snapshot has been taken" >&2
  echo "The interval defines the time to sleep in seconds between replications. An interval 0 means the script will not loop and just do one replication." >&2
  echo "The retries defines how many retries should be made before the script fails (default: 5)" >&2
  echo "The keepsnaps defines how many snapshots should be kept" >&2
  echo "The ssh_privatekeyfile is the private key file used to connect to the user@host without password" >&2
  cleanup 1
}

before_snap() {
  if [ -n "$beforescript" ] ; then 
    log "I" "Running before script $beforescript"
    $beforescript
    status=$?
    if [ $status -ne 0 ] ; then
      log "E" "The before script $beforescript failed with exit code $status"
      return 1
    fi
    log "I" "End of before script $beforescript"
  fi
}
after_snap(){
  if [ -n "$afterscript" ] ; then
    log "I" "Running after script $afterscript"
    $afterscript
    status=$?
    if [ $status -ne 0 ] ; then
      log "E" "The after script $afterscript failed with exit code $status"
      return 1
    fi
    log "I" "End of after script $afterscript"
  fi
}
do_src_snap() {
   zfs snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$1 
   status=$?
   if [ $status -ne 0 ] ; then
     log "E" "Failed to create snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$1  with exit code $status"
     return 1
   fi
}

connectssh() {
  ssh -o "BatchMode yes" -o "StrictHostKeyChecking no" -i $privatekey $connect $* 
}
sourcemaxindex=0
destmaxindex=0
while getopts s:d:c:i:b:a:k:p:Cr: option ; do
  case $option in
    s) let sourcemaxindex++ ; source_zfs[$sourcemaxindex]=$OPTARG ;;
    d) let destmaxindex++ ; dest_zfs[$destmaxindex]=$OPTARG ;;
    c) connect=$OPTARG ;;
    i) interval=$OPTARG ;;
    b) beforescript=$OPTARG ;;
    a) afterscript=$OPTARG ;;
    k) keepsnaps=$OPTARG ;;
    p) privatekey=$OPTARG ;;
    C) checkonly=true ;;
    r) MAXRETRIES=$OPTARG ;;
    *) usage ;;
  esac
done

if [ -n "${keepsnaps##[0-9]*}" -o -z "$keepsnaps" ] ; then 
  keepsnaps=1
fi
if [ $keepsnaps -lt 1 ] ; then
  keepsnaps=1
fi
if [ -z "$privatekey" ] ; then
  log "E" "No private key file specified"
  cleanup 10
fi
if [ $sourcemaxindex -eq 0 -o $sourcemaxindex -ne $destmaxindex -o -z "$connect" -o -z "$interval" ] ; then
  usage
fi

index=0
while [ $index -lt $sourcemaxindex ] ; do
  let index++
  log "I" "Check if the source filesystem exists"
  if ! zfs list ${source_zfs[$index]} >/dev/null 2>&1 ; then
    log E "The source ${source_zfs[$index]} zfs filesystem does not exist!" >&2
    cleanup 2
  fi
done

retries=0
status=0
log "I" "Trying to connect without password"
if ! connectssh "exit 0" >/dev/null 2>&1 ; then
  log E "Cannot connect to $connect without password."
  cleanup 3
fi
retries=0
status=0

index=0
while [ $index -lt $sourcemaxindex ] ; do
  let index++
  log "I" "Checking the destination filesystem"
  while ! connectssh "/usr/sbin/zfs list `dirname ${dest_zfs[$index]}`" >/dev/null 2>&1 ; do 
    log W "zfs filesystem ${dest_zfs[$index]} is not a valid destination, because parent filesystem does not exist. Retrying..." 
    if [ $retries -eq $MAXRETRIES ] ; then
      status=1
      break
    fi
    sleep 5
  done
  if [ $status -ne 0 ] ; then
    log E "zfs filesystem ${dest_zfs[$index]} is not a valid destination, because parent filesystem does not exist" 
    cleanup 4
  fi
done
# Todo check allow for filesystems here
index=0
while [ $index -lt $sourcemaxindex ] ; do
  let index++
  log "I" "Checking the permissions on the source filesystem"
  user=`id | sed 's/[^(]*(\([^)]*\).*/\1/g'`
  srcuser=$user
  zfs allow ${source_zfs[$index]} > $TMPDIR/zfsallow.out 2>/dev/null
  if ! grep "user $user .*send.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to send ${source_zfs[$index]} 
    cleanup 5
  fi
  if ! grep "user $user .*destroy.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to destroy ${source_zfs[$index]} 
    cleanup 5
  fi
  if ! grep "user $user .*mount.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to mount ${source_zfs[$index]} 
    cleanup 5
  fi
  if ! grep "user $user .*snapshot.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to snapshot ${source_zfs[$index]} 
    cleanup 5
  fi
  if ! grep "user $user .*create.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to create ${source_zfs[$index]} 
    cleanup 5
  fi
done


index=0
while [ $index -lt $sourcemaxindex ] ; do
  let index++
  log "I" "Checking the permissions on the destination filesystem"
  user=${connect%@*}
  dstuser=$user
  if [ -z "$dsthost" ] ; then
    dsthost=${connect#*@}
  fi
  retries=0
  status=0
  while ! connectssh "/usr/sbin/zfs allow `dirname ${dest_zfs[$index]}`" > $TMPDIR/zfsallow.out 2>/dev/null ; do
    log "W" "Cannot get the permission on destination filesystem. Retrying..."
    if [ $retries -eq $MAXRETRIES ] ; then
      status=1
      break
    fi
    sleep 5
  done
  if [ $status -ne 0 ] ; then
    log E "Cannot get the permission on destination filesystem."
    cleanup 4
  fi
  if ! grep "user $user .*create.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to create ${dest_zfs[$index]} 
    cleanup 5
  fi
  if ! grep "user $user .*destroy.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to destroy ${dest_zfs[$index]} 
    cleanup 5
  fi
  if ! grep "user $user .*mount.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to mount ${dest_zfs[$index]} 
    cleanup 5
  fi
  if ! grep "user $user .*receive.*" $TMPDIR/zfsallow.out >/dev/null 2>/dev/null; then
    log E user $user is not allowed to receive ${dest_zfs[$index]} 
    cleanup 5
  fi
done

if [ "$checkonly" = "true" ] ; then
  cleanup 0
fi
# Check last snapshot which was transfered 

if [ "$remote" = "true" ] ; then
  tmpuser=$dstuser
  dstuser=$srcuser
  srcuser=$tmpuser
fi

index=0
while [ $index -lt $sourcemaxindex ] ; do
  let index++
  lastsnap[$index]=`zfs list -Ho name -t snapshot|nawk -F @ -v source_zfs=${source_zfs[$index]} '$1==source_zfs {print $2}'|nawk -F- -v srcuser=$srcuser -v dstuser=$dstuser -v dsthost=$dsthost '$1=="zfsrepl" && $2==srcuser && $3==dstuser && $4==dsthost {print $5}'|sort -n|tail -1`
  # Transfer first snapshot if no snapshot has been transfered yet
  if [ -z "${lastsnap[$index]}" ] ; then
    log "I" "Creating initial snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-0"
    do_src_snap 0
    status=$?
    if [ $status -ne 0 ] ; then
      log "E" "The snapshot creation failed"
      cleanup 5
    fi
    lastsnap[$index]=0
  fi
  if [ ${lastsnap[$index]} -eq 0 ] ; then
    if ! connectssh "/usr/sbin/zfs list ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-0 1>/dev/null 2>&1" ; then 
      log "I" "Sending initial snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-0"
      zfs send -p ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-0|connectssh "/usr/sbin/zfs receive ${dest_zfs[$index]}"
      status=$?
      if [ $status -ne 0 ] ; then
        log "E" "Sending initial snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-0 failed with exit code $status"
        cleanup 5
      fi
    fi
  fi
  # Cleanup snapshots which are to old
  firstsnap[$index]=`zfs list -Ho name -t snapshot|nawk -F @ -v source_zfs=${source_zfs[$index]} '$1==source_zfs {print $2}'|nawk -F- -v srcuser=$srcuser -v dstuser=$dstuser -v dsthost=$dsthost '$1=="zfsrepl" && $2==srcuser && $3==dstuser && $4==dsthost {print $5}'|sort -n|head -1`
  if [ $((${lastsnap[$index]} - ${firstsnap[$index]} - 1)) -ge $keepsnaps ] ; then 
    i=${firstsnap[$index]}
    while [ $((${lastsnap[$index]} - $i - 1)) -ge $keepsnaps ] ; do
      retries=0
      log "I" "Destroying ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i"
      while ! (zfs list ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i >/dev/null 2>&1 && zfs destroy ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i || exit 0); do 
        sleep 5
        let retries++
        log "W" "Retry destroying ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i" 
        if [ $retries -eq $MAXRETRIES ] ; then
          break
        fi
      done
      retries=0
      log "I" "Destroying ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i"
      while ! connectssh "/usr/sbin/zfs list ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i >/dev/null 2>&1 && /usr/sbin/zfs destroy ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i || exit 0" ; do 
        sleep 5
        let retries++
        log "W" "Retry destroying ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$i" 
        if [ $retries -eq $MAXRETRIES ] ; then
          break
        fi
      done
      let i++
    done
  fi 
done

# Loop and tranfer incremental snapshots 

count=0
while [ $interval -ne 0 -o $count -le 1 ] ; do
  index=0
  while [ $index -lt $sourcemaxindex ] ; do
    let index++
    retries=0
    while [ $retries -lt $MAXRETRIES ] ; do
      if ! connectssh "/usr/sbin/zfs list ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-${lastsnap[$index]}" > $TMPDIR/zfslist.out 2>&1 ; then
        if grep "does not exist" $TMPDIR/zfslist.out 1>/dev/null ; then
          log "I" "Sending ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-${lastsnap[$index]}"
          zfs send -i ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-1)) ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-${lastsnap[$index]} |connectssh "/usr/sbin/zfs receive -F ${dest_zfs[$index]}"
          status=$?
          if [ $status -ne 0 ] ; then
            log "W" "Sending the incremental snapshot  ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-${lastsnap[$index]} failed with exit code $status"
          else
            break
          fi
        fi
      else
        break
      fi
      let retries++
      sleep 5
    done
    if [ $MAXRETRIES -eq $retries ] ; then
      do_alarming "Could not send snapshots even after $retries retries. Stopping here now"
      cleanup 7
    fi
  done
  index=0
  while [ $index -lt $sourcemaxindex ] ; do
    let index++
    retries=0
    status=0
    if [ $((${lastsnap[$index]}-$keepsnaps)) -ge 0 ] ; then
      log "I" "Destroying ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps))"
      while ! (zfs list ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps)) >/dev/null 2>&1 && zfs destroy ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps)) || exit 0); do 
        sleep 5
        let retries++
        log "W" "Retry destroying ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps))" 
        if [ $retries -eq $MAXRETRIES ] ; then
          status=1
          break
        fi
      done
      if [ $status -ne 0 ] ; then
        log "E" "Destroying snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps)) failed"
        cleanup 6
      fi
      retries=0
      status=0
      log "I" "Destroying ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps))"
      while ! connectssh "/usr/sbin/zfs list ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps)) >/dev/null 2>&1 && /usr/sbin/zfs destroy ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps)) || exit 0" ; do 
        sleep 5
        let retries++
        log "W" "Retry destroying ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps))" 
        if [ $retries -eq $MAXRETRIES ] ; then
          status=1
          break
        fi
      done
      if [ $status -ne 0 ] ; then
        log "E" "Destroying snapshot ${dest_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-$((${lastsnap[$index]}-$keepsnaps)) failed"
        cleanup 6
      fi
    fi
    let lastsnap[$index]++
  done
  sleep $interval
  if [ $interval -eq 0 -a $count -eq 1 ] ; then 
    break
  fi
  if ! before_snap ; then
     cleanup 11
  fi 
  index=0
  while [ $index -lt $sourcemaxindex ] ; do
    let index++
    retries=0
    status=0
    log "I" "Creating snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-${lastsnap[$index]}"
    while ! do_src_snap ${lastsnap[$index]} ; do 
      sleep 5
      let retries++
      log "W" "Retry creating snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-${lastsnap[$index]}"
      if [ $retries -eq $MAXRETRIES ] ; then
        status=1
        break
      fi
    done
    if [ $status -ne 0 ] ; then
      log "E" "Creating snapshot ${source_zfs[$index]}@zfsrepl-$srcuser-$dstuser-$dsthost-${lastsnap[$index]} failed"
      cleanup 8
    fi
  done
  if ! after_snap ; then
     cleanup 11
  fi 
  let count++
done
cleanup 0
